@@ -1268,6 +1268,7 @@ func TestAssertTypeJSONWrongLoader(t *testing.T) {
 		files: map[string]string{
 			"/entry.js": `
 				import foo from './foo.json' assert { type: 'json' }
+				console.log(foo)
 			`,
 			"/foo.json": `{}`,
 		},
@@ -1286,6 +1287,62 @@ NOTE: You need to either reconfigure esbuild to ensure that the loader for this
 	})
 }
 
+func TestWithTypeJSONOverrideLoader(t *testing.T) {
+	loader_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.js": `
+				import foo from './foo.js' with { type: 'json' }
+				console.log(foo)
+			`,
+			"/foo.js": `{ "this is json not js": true }`,
+		},
+		entryPaths: []string{"/entry.js"},
+		options: config.Options{
+			Mode: config.ModeBundle,
+		},
+	})
+}
+
+func TestWithBadType(t *testing.T) {
+	loader_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.js": `
+				import foo from './foo.json' with { type: '' }
+				import bar from './foo.json' with { type: 'garbage' }
+				console.log(bar)
+			`,
+			"/foo.json": `{}`,
+		},
+		entryPaths: []string{"/entry.js"},
+		options: config.Options{
+			Mode: config.ModeBundle,
+		},
+		expectedScanLog: `entry.js: ERROR: Importing with a type attribute of "" is not supported
+entry.js: ERROR: Importing with a type attribute of "garbage" is not supported
+`,
+	})
+}
+
+func TestWithBadAttribute(t *testing.T) {
+	loader_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.js": `
+				import foo from './foo.json' with { '': 'json' }
+				import bar from './foo.json' with { garbage: 'json' }
+				console.log(bar)
+			`,
+			"/foo.json": `{}`,
+		},
+		entryPaths: []string{"/entry.js"},
+		options: config.Options{
+			Mode: config.ModeBundle,
+		},
+		expectedScanLog: `entry.js: ERROR: Importing with the "" attribute is not supported
+entry.js: ERROR: Importing with the "garbage" attribute is not supported
+`,
+	})
+}
+
 func TestEmptyLoaderJS(t *testing.T) {
 	loader_suite.expectBundled(t, bundled{
 		files: map[string]string{
@@ -1537,3 +1594,41 @@ func TestLoaderCopyWithInjectedFileBundle(t *testing.T) {
 		},
 	})
 }
+
+func TestLoaderBundleWithImportAttributes(t *testing.T) {
+	loader_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.js": `
+				import x from "./data.json"
+				import y from "./data.json" assert { type: 'json' }
+				import z from "./data.json" with { type: 'json' }
+				console.log(x === y, x !== z)
+			`,
+			"/data.json": `{ "works": true }`,
+		},
+		entryPaths: []string{"/entry.js"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.js",
+		},
+	})
+}
+
+func TestLoaderBundleWithTypeJSONOnlyDefaultExport(t *testing.T) {
+	loader_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.js": `
+				import x, {foo as x2} from "./data.json"
+				import y, {foo as y2} from "./data.json" with { type: 'json' }
+			`,
+			"/data.json": `{ "foo": 123 }`,
+		},
+		entryPaths: []string{"/entry.js"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.js",
+		},
+		expectedCompileLog: `entry.js: ERROR: No matching export in "data.json with { type: 'json' }" for import "foo"
+`,
+	})
+}
