@@ -1,6 +1,7 @@
 package bundler_tests
 
 import (
+	"regexp"
 	"testing"
 
 	"github.com/evanw/esbuild/internal/compat"
@@ -128,9 +129,9 @@ func TestTSDeclareClassFields(t *testing.T) {
 				}
 				(() => new Bar())()
 			`,
-			"/define-true/tsconfig.json": `{
+			"/define-false/tsconfig.json": `{
 				"compilerOptions": {
-					"useDefineForClassFields": true
+					"useDefineForClassFields": false
 				}
 			}`,
 		},
@@ -422,6 +423,47 @@ func TestTSExportNamespace(t *testing.T) {
 	})
 }
 
+func TestTSNamespaceKeepNames(t *testing.T) {
+	ts_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.ts": `
+				namespace ns {
+					export let foo = () => {}
+					export function bar() {}
+					export class Baz {}
+				}
+			`,
+		},
+		entryPaths: []string{"/entry.ts"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.js",
+			KeepNames:     true,
+		},
+	})
+}
+
+func TestTSNamespaceKeepNamesTargetES2015(t *testing.T) {
+	ts_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.ts": `
+				namespace ns {
+					export let foo = () => {}
+					export function bar() {}
+					export class Baz {}
+				}
+			`,
+		},
+		entryPaths: []string{"/entry.ts"},
+		options: config.Options{
+			Mode:                  config.ModeBundle,
+			AbsOutputFile:         "/out.js",
+			KeepNames:             true,
+			UnsupportedJSFeatures: es(2015),
+		},
+	})
+}
+
 func TestTSMinifyEnum(t *testing.T) {
 	ts_suite.expectBundled(t, bundled{
 		files: map[string]string{
@@ -604,6 +646,11 @@ func TestTSMinifyDerivedClass(t *testing.T) {
 					}
 				}
 			`,
+			"/tsconfig.json": `{
+				"compilerOptions": {
+					"useDefineForClassFields": false
+				}
+			}`,
 		},
 		entryPaths: []string{"/entry.ts"},
 		options: config.Options{
@@ -614,6 +661,110 @@ func TestTSMinifyDerivedClass(t *testing.T) {
 	})
 }
 
+func TestTSMinifyEnumPropertyNames(t *testing.T) {
+	ts_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.ts": `
+				import { CrossFileGood, CrossFileBad } from './cross-file'
+				const enum SameFileGood {
+					STR = 'str 1',
+					NUM = 123,
+				}
+				const enum SameFileBad {
+					PROTO = '__proto__',
+					CONSTRUCTOR = 'constructor',
+					PROTOTYPE = 'prototype',
+				}
+				class Foo {
+					[100] = 100;
+					'200' = 200;
+					['300'] = 300;
+					[SameFileGood.STR] = SameFileGood.STR;
+					[SameFileGood.NUM] = SameFileGood.NUM;
+					[CrossFileGood.STR] = CrossFileGood.STR;
+					[CrossFileGood.NUM] = CrossFileGood.NUM;
+				}
+				shouldNotBeComputed(
+					class {
+						[100] = 100;
+						'200' = 200;
+						['300'] = 300;
+						[SameFileGood.STR] = SameFileGood.STR;
+						[SameFileGood.NUM] = SameFileGood.NUM;
+						[CrossFileGood.STR] = CrossFileGood.STR;
+						[CrossFileGood.NUM] = CrossFileGood.NUM;
+					},
+					{
+						[100]: 100,
+						'200': 200,
+						['300']: 300,
+						[SameFileGood.STR]: SameFileGood.STR,
+						[SameFileGood.NUM]: SameFileGood.NUM,
+						[CrossFileGood.STR]: CrossFileGood.STR,
+						[CrossFileGood.NUM]: CrossFileGood.NUM,
+					},
+				)
+				mustBeComputed(
+					{ [SameFileBad.PROTO]: null },
+					{ [CrossFileBad.PROTO]: null },
+					class { [SameFileBad.CONSTRUCTOR]() {} },
+					class { [CrossFileBad.CONSTRUCTOR]() {} },
+					class { static [SameFileBad.PROTOTYPE]() {} },
+					class { static [CrossFileBad.PROTOTYPE]() {} },
+				)
+			`,
+			"/cross-file.ts": `
+				export const enum CrossFileGood {
+					STR = 'str 2',
+					NUM = 321,
+				}
+				export const enum CrossFileBad {
+					PROTO = '__proto__',
+					CONSTRUCTOR = 'constructor',
+					PROTOTYPE = 'prototype',
+				}
+			`,
+		},
+		entryPaths: []string{"/entry.ts"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			MinifySyntax:  true,
+			AbsOutputFile: "/out.js",
+		},
+	})
+}
+func TestTSMinifyEnumCrossFileInlineStringsIntoTemplates(t *testing.T) {
+	ts_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.ts": `
+				import { CrossFile } from './cross-file'
+				enum SameFile {
+					STR = 'str 1',
+					NUM = 123,
+				}
+				console.log(` + "`" + `
+					SameFile.STR = ${SameFile.STR}
+					SameFile.NUM = ${SameFile.NUM}
+					CrossFile.STR = ${CrossFile.STR}
+					CrossFile.NUM = ${CrossFile.NUM}
+				` + "`" + `)
+			`,
+			"/cross-file.ts": `
+				export enum CrossFile {
+					STR = 'str 2',
+					NUM = 321,
+				}
+			`,
+		},
+		entryPaths: []string{"/entry.ts"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			MinifySyntax:  true,
+			AbsOutputFile: "/out.js",
+		},
+	})
+}
+
 func TestTSImportVsLocalCollisionAllTypes(t *testing.T) {
 	ts_suite.expectBundled(t, bundled{
 		files: map[string]string{
@@ -747,6 +898,36 @@ func TestTSImportEqualsBundle(t *testing.T) {
 	})
 }
 
+func TestTSImportEqualsUndefinedImport(t *testing.T) {
+	ts_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.ts": `
+				import * as ns from './import.ts'
+				import value_copy = ns.value
+				import Type_copy = ns.Type
+				let foo: Type_copy = value_copy
+				console.log(foo)
+			`,
+			"/import.ts": `
+				export let value = 123
+				export type Type = number
+			`,
+		},
+		entryPaths: []string{"/entry.ts"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.js",
+			ExternalSettings: config.ExternalSettings{
+				PreResolve: config.ExternalMatchers{
+					Exact: map[string]bool{
+						"pkg": true,
+					},
+				},
+			},
+		},
+	})
+}
+
 func TestTSMinifiedBundleES6(t *testing.T) {
 	ts_suite.expectBundled(t, bundled{
 		files: map[string]string{
@@ -798,7 +979,54 @@ func TestTSMinifiedBundleCommonJS(t *testing.T) {
 	})
 }
 
-func TestTypeScriptDecorators(t *testing.T) {
+func TestTSExperimentalDecoratorsNoConfig(t *testing.T) {
+	ts_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.ts": `
+				declare let x: any, y: any
+				@x.y()
+				@(new y.x)
+				export default class Foo {
+					@x @y mUndef: any
+					@x @y mDef = 1
+					@x @y method() { return new Foo }
+					@x @y accessor aUndef: any
+					@x @y accessor aDef = 1
+
+					@x @y static sUndef: any
+					@x @y static sDef = new Foo
+					@x @y static sMethod() { return new Foo }
+					@x @y static accessor asUndef: any
+					@x @y static accessor asDef = 1
+
+					@x @y #mUndef: any
+					@x @y #mDef = 1
+					@x @y #method() { return new Foo }
+					@x @y accessor #aUndef: any
+					@x @y accessor #aDef = 1
+
+					@x @y static #sUndef: any
+					@x @y static #sDef = 1
+					@x @y static #sMethod() { return new Foo }
+					@x @y static accessor #asUndef: any
+					@x @y static accessor #asDef = 1
+				}
+			`,
+			"/tsconfig.json": `{
+				"compilerOptions": {
+					"experimentalDecorators": false
+				}
+			}`,
+		},
+		entryPaths: []string{"/entry.ts"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.js",
+		},
+	})
+}
+
+func TestTSExperimentalDecorators(t *testing.T) {
 	ts_suite.expectBundled(t, bundled{
 		files: map[string]string{
 			"/entry.js": `
@@ -935,6 +1163,12 @@ func TestTypeScriptDecorators(t *testing.T) {
 					return Foo;
 				}
 			`,
+			"/tsconfig.json": `{
+				"compilerOptions": {
+					"useDefineForClassFields": false,
+					"experimentalDecorators": true
+				}
+			}`,
 		},
 		entryPaths: []string{"/entry.js"},
 		options: config.Options{
@@ -944,13 +1178,18 @@ func TestTypeScriptDecorators(t *testing.T) {
 	})
 }
 
-func TestTypeScriptDecoratorsKeepNames(t *testing.T) {
+func TestTSExperimentalDecoratorsKeepNames(t *testing.T) {
 	ts_suite.expectBundled(t, bundled{
 		files: map[string]string{
 			"/entry.ts": `
 				@decoratorMustComeAfterName
-				class Foo {}
+				export class Foo {}
 			`,
+			"/tsconfig.json": `{
+				"compilerOptions": {
+					"experimentalDecorators": true
+				}
+			}`,
 		},
 		entryPaths: []string{"/entry.ts"},
 		options: config.Options{
@@ -962,7 +1201,7 @@ func TestTypeScriptDecoratorsKeepNames(t *testing.T) {
 }
 
 // See: https://github.com/evanw/esbuild/issues/2147
-func TestTypeScriptDecoratorScopeIssue2147(t *testing.T) {
+func TestTSExperimentalDecoratorScopeIssue2147(t *testing.T) {
 	ts_suite.expectBundled(t, bundled{
 		files: map[string]string{
 			"/entry.ts": `
@@ -989,6 +1228,12 @@ func TestTypeScriptDecoratorScopeIssue2147(t *testing.T) {
 					}
 				}
 			`,
+			"/tsconfig.json": `{
+				"compilerOptions": {
+					"useDefineForClassFields": false,
+					"experimentalDecorators": true
+				}
+			}`,
 		},
 		entryPaths: []string{"/entry.ts"},
 		options: config.Options{
@@ -1117,6 +1362,11 @@ func TestTSExportDefaultTypeIssue316(t *testing.T) {
 				export default foo
 				export let bar = 123
 			`,
+			"/tsconfig.json": `{
+				"compilerOptions": {
+					"useDefineForClassFields": false
+				}
+			}`,
 		},
 		entryPaths: []string{"/entry.ts"},
 		options: config.Options{
@@ -1136,6 +1386,14 @@ func TestTSImplicitExtensions(t *testing.T) {
 				import './pick-tsx.jsx'
 				import './order-js.js'
 				import './order-jsx.jsx'
+
+				import 'pkg/foo-js.js'
+				import 'pkg/foo-jsx.jsx'
+				import 'pkg-exports/xyz-js'
+				import 'pkg-exports/xyz-jsx'
+				import 'pkg-exports/foo-js.js'
+				import 'pkg-exports/foo-jsx.jsx'
+				import 'pkg-imports'
 			`,
 
 			"/pick-js.js": `console.log("correct")`,
@@ -1155,6 +1413,39 @@ func TestTSImplicitExtensions(t *testing.T) {
 
 			"/order-jsx.ts":  `console.log("correct")`,
 			"/order-jsx.tsx": `console.log("wrong")`,
+
+			"/node_modules/pkg/foo-js.ts":   `console.log("correct")`,
+			"/node_modules/pkg/foo-jsx.tsx": `console.log("correct")`,
+
+			"/node_modules/pkg-exports/package.json": `{
+				"exports": {
+					"./xyz-js": "./abc-js.js",
+					"./xyz-jsx": "./abc-jsx.jsx",
+					"./*": "./lib/*"
+				}
+			}`,
+			"/node_modules/pkg-exports/abc-js.ts":       `console.log("correct")`,
+			"/node_modules/pkg-exports/abc-jsx.tsx":     `console.log("correct")`,
+			"/node_modules/pkg-exports/lib/foo-js.ts":   `console.log("correct")`,
+			"/node_modules/pkg-exports/lib/foo-jsx.tsx": `console.log("correct")`,
+
+			"/node_modules/pkg-imports/package.json": `{
+				"imports": {
+					"#xyz-js": "./abc-js.js",
+					"#xyz-jsx": "./abc-jsx.jsx",
+					"#bar/*": "./lib/*"
+				}
+			}`,
+			"/node_modules/pkg-imports/index.js": `
+				import "#xyz-js"
+				import "#xyz-jsx"
+				import "#bar/foo-js.js"
+				import "#bar/foo-jsx.jsx"
+			`,
+			"/node_modules/pkg-imports/abc-js.ts":       `console.log("correct")`,
+			"/node_modules/pkg-imports/abc-jsx.tsx":     `console.log("correct")`,
+			"/node_modules/pkg-imports/lib/foo-js.ts":   `console.log("correct")`,
+			"/node_modules/pkg-imports/lib/foo-jsx.tsx": `console.log("correct")`,
 		},
 		entryPaths: []string{"/entry.ts"},
 		options: config.Options{
@@ -1230,9 +1521,11 @@ func TestExportTypeIssue379(t *testing.T) {
 		},
 		entryPaths: []string{"/entry.ts"},
 		options: config.Options{
-			Mode:                    config.ModeBundle,
-			AbsOutputFile:           "/out.js",
-			UseDefineForClassFields: config.False,
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.js",
+			TS: config.TSOptions{Config: config.TSConfig{
+				UseDefineForClassFields: config.False,
+			}},
 		},
 	})
 }
@@ -1269,9 +1562,11 @@ func TestThisInsideFunctionTS(t *testing.T) {
 		},
 		entryPaths: []string{"/entry.ts"},
 		options: config.Options{
-			Mode:                    config.ModeBundle,
-			AbsOutputFile:           "/out.js",
-			UseDefineForClassFields: config.False,
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.js",
+			TS: config.TSOptions{Config: config.TSConfig{
+				UseDefineForClassFields: config.False,
+			}},
 		},
 	})
 }
@@ -1308,9 +1603,11 @@ func TestThisInsideFunctionTSUseDefineForClassFields(t *testing.T) {
 		},
 		entryPaths: []string{"/entry.ts"},
 		options: config.Options{
-			Mode:                    config.ModeBundle,
-			AbsOutputFile:           "/out.js",
-			UseDefineForClassFields: config.True,
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.js",
+			TS: config.TSOptions{Config: config.TSConfig{
+				UseDefineForClassFields: config.True,
+			}},
 		},
 	})
 }
@@ -1344,6 +1641,11 @@ func TestThisInsideFunctionTSNoBundle(t *testing.T) {
 					new Bar(bar(objBar))
 				}
 			`,
+			"/tsconfig.json": `{
+				"compilerOptions": {
+					"useDefineForClassFields": false
+				}
+			}`,
 		},
 		entryPaths: []string{"/entry.ts"},
 		options: config.Options{
@@ -1385,9 +1687,11 @@ func TestThisInsideFunctionTSNoBundleUseDefineForClassFields(t *testing.T) {
 		},
 		entryPaths: []string{"/entry.ts"},
 		options: config.Options{
-			Mode:                    config.ModePassThrough,
-			AbsOutputFile:           "/out.js",
-			UseDefineForClassFields: config.True,
+			Mode:          config.ModePassThrough,
+			AbsOutputFile: "/out.js",
+			TS: config.TSOptions{Config: config.TSConfig{
+				UseDefineForClassFields: config.True,
+			}},
 		},
 	})
 }
@@ -1406,12 +1710,17 @@ func TestTSComputedClassFieldUseDefineFalse(t *testing.T) {
 				}
 				new Foo()
 			`,
+			"/tsconfig.json": `{
+				"compilerOptions": {
+					"useDefineForClassFields": false,
+					"experimentalDecorators": true
+				}
+			}`,
 		},
 		entryPaths: []string{"/entry.ts"},
 		options: config.Options{
-			Mode:                    config.ModePassThrough,
-			AbsOutputFile:           "/out.js",
-			UseDefineForClassFields: config.False,
+			Mode:          config.ModePassThrough,
+			AbsOutputFile: "/out.js",
 		},
 	})
 }
@@ -1430,12 +1739,17 @@ func TestTSComputedClassFieldUseDefineTrue(t *testing.T) {
 				}
 				new Foo()
 			`,
+			"/tsconfig.json": `{
+				"compilerOptions": {
+					"useDefineForClassFields": true,
+					"experimentalDecorators": true
+				}
+			}`,
 		},
 		entryPaths: []string{"/entry.ts"},
 		options: config.Options{
-			Mode:                    config.ModePassThrough,
-			AbsOutputFile:           "/out.js",
-			UseDefineForClassFields: config.True,
+			Mode:          config.ModePassThrough,
+			AbsOutputFile: "/out.js",
 		},
 	})
 }
@@ -1454,13 +1768,18 @@ func TestTSComputedClassFieldUseDefineTrueLower(t *testing.T) {
 				}
 				new Foo()
 			`,
+			"/tsconfig.json": `{
+				"compilerOptions": {
+					"useDefineForClassFields": true,
+					"experimentalDecorators": true
+				}
+			}`,
 		},
 		entryPaths: []string{"/entry.ts"},
 		options: config.Options{
-			Mode:                    config.ModePassThrough,
-			AbsOutputFile:           "/out.js",
-			UseDefineForClassFields: config.True,
-			UnsupportedJSFeatures:   compat.ClassField,
+			Mode:                  config.ModePassThrough,
+			AbsOutputFile:         "/out.js",
+			UnsupportedJSFeatures: compat.ClassField,
 		},
 	})
 }
@@ -1483,9 +1802,11 @@ func TestTSAbstractClassFieldUseAssign(t *testing.T) {
 		},
 		entryPaths: []string{"/entry.ts"},
 		options: config.Options{
-			Mode:                    config.ModePassThrough,
-			AbsOutputFile:           "/out.js",
-			UseDefineForClassFields: config.False,
+			Mode:          config.ModePassThrough,
+			AbsOutputFile: "/out.js",
+			TS: config.TSOptions{Config: config.TSConfig{
+				UseDefineForClassFields: config.False,
+			}},
 		},
 	})
 }
@@ -1508,9 +1829,11 @@ func TestTSAbstractClassFieldUseDefine(t *testing.T) {
 		},
 		entryPaths: []string{"/entry.ts"},
 		options: config.Options{
-			Mode:                    config.ModePassThrough,
-			AbsOutputFile:           "/out.js",
-			UseDefineForClassFields: config.True,
+			Mode:          config.ModePassThrough,
+			AbsOutputFile: "/out.js",
+			TS: config.TSOptions{Config: config.TSConfig{
+				UseDefineForClassFields: config.True,
+			}},
 		},
 	})
 }
@@ -1830,18 +2153,35 @@ func TestTSEnumSameModuleInliningAccess(t *testing.T) {
 	ts_suite.expectBundled(t, bundled{
 		files: map[string]string{
 			"/entry.ts": `
-				enum a { x = 123 }
-				enum b { x = 123 }
-				enum c { x = 123 }
-				enum d { x = 123 }
-				enum e { x = 123 }
-				console.log([
-					a.x,
-					b['x'],
-					c?.x,
-					d?.['x'],
-					e,
-				])
+				enum a_num { x = 123 }
+				enum b_num { x = 123 }
+				enum c_num { x = 123 }
+				enum d_num { x = 123 }
+				enum e_num { x = 123 }
+
+				enum a_str { x = 'abc' }
+				enum b_str { x = 'abc' }
+				enum c_str { x = 'abc' }
+				enum d_str { x = 'abc' }
+				enum e_str { x = 'abc' }
+
+				inlined = [
+					a_num.x,
+					b_num['x'],
+
+					a_str.x,
+					b_str['x'],
+				]
+
+				not_inlined = [
+					c_num?.x,
+					d_num?.['x'],
+					e_num,
+
+					c_str?.x,
+					d_str?.['x'],
+					e_str,
+				]
 			`,
 		},
 		entryPaths: []string{"/entry.ts"},
@@ -1856,21 +2196,41 @@ func TestTSEnumCrossModuleInliningAccess(t *testing.T) {
 	ts_suite.expectBundled(t, bundled{
 		files: map[string]string{
 			"/entry.ts": `
-				import { a, b, c, d, e } from './enums'
-				console.log([
-					a.x,
-					b['x'],
-					c?.x,
-					d?.['x'],
-					e,
-				])
+				import {
+					a_num, b_num, c_num, d_num, e_num,
+					a_str, b_str, c_str, d_str, e_str,
+				} from './enums'
+
+				inlined = [
+					a_num.x,
+					b_num['x'],
+
+					a_str.x,
+					b_str['x'],
+				]
+
+				not_inlined = [
+					c_num?.x,
+					d_num?.['x'],
+					e_num,
+
+					c_str?.x,
+					d_str?.['x'],
+					e_str,
+				]
 			`,
 			"/enums.ts": `
-				export enum a { x = 123 }
-				export enum b { x = 123 }
-				export enum c { x = 123 }
-				export enum d { x = 123 }
-				export enum e { x = 123 }
+				export enum a_num { x = 123 }
+				export enum b_num { x = 123 }
+				export enum c_num { x = 123 }
+				export enum d_num { x = 123 }
+				export enum e_num { x = 123 }
+
+				export enum a_str { x = 'abc' }
+				export enum b_str { x = 'abc' }
+				export enum c_str { x = 'abc' }
+				export enum d_str { x = 'abc' }
+				export enum e_str { x = 'abc' }
 			`,
 		},
 		entryPaths: []string{"/entry.ts"},
@@ -1943,6 +2303,48 @@ func TestTSEnumCrossModuleInliningReExport(t *testing.T) {
 	})
 }
 
+func TestTSEnumCrossModuleInliningMinifyIndexIntoDot(t *testing.T) {
+	ts_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.ts": `
+				const enum Foo {
+					foo1 = 'abc',
+					foo2 = 'a b c',
+				}
+				import { Bar } from './lib'
+				inlined = [
+					obj[Foo.foo1],
+					obj[Bar.bar1],
+					obj?.[Foo.foo1],
+					obj?.[Bar.bar1],
+					obj?.prop[Foo.foo1],
+					obj?.prop[Bar.bar1],
+				]
+				notInlined = [
+					obj[Foo.foo2],
+					obj[Bar.bar2],
+					obj?.[Foo.foo2],
+					obj?.[Bar.bar2],
+					obj?.prop[Foo.foo2],
+					obj?.prop[Bar.bar2],
+				]
+			`,
+			"/lib.ts": `
+				export const enum Bar {
+					bar1 = 'xyz',
+					bar2 = 'x y z',
+				}
+			`,
+		},
+		entryPaths: []string{"/entry.ts"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.js",
+			MinifySyntax:  true,
+		},
+	})
+}
+
 func TestTSEnumCrossModuleTreeShaking(t *testing.T) {
 	ts_suite.expectBundled(t, bundled{
 		files: map[string]string{
@@ -2280,3 +2682,301 @@ func TestTSEnumUseBeforeDeclare(t *testing.T) {
 		},
 	})
 }
+
+func TestTSPreferJSOverTSInsideNodeModules(t *testing.T) {
+	// We now prefer ".js" over ".ts" inside "node_modules"
+	ts_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/Users/user/project/src/main.ts": `
+				// Implicit extensions
+				import './relative/path'
+				import 'package/path'
+
+				// Explicit extensions
+				import './relative2/path.js'
+				import 'package2/path.js'
+			`,
+
+			"/Users/user/project/src/relative/path.ts": `console.log('success')`,
+			"/Users/user/project/src/relative/path.js": `console.log('FAILURE')`,
+
+			"/Users/user/project/src/relative2/path.ts": `console.log('FAILURE')`,
+			"/Users/user/project/src/relative2/path.js": `console.log('success')`,
+
+			"/Users/user/project/node_modules/package/path.ts": `console.log('FAILURE')`,
+			"/Users/user/project/node_modules/package/path.js": `console.log('success')`,
+
+			"/Users/user/project/node_modules/package2/path.ts": `console.log('FAILURE')`,
+			"/Users/user/project/node_modules/package2/path.js": `console.log('success')`,
+		},
+		entryPaths: []string{"/Users/user/project/src/main.ts"},
+		options: config.Options{
+			Mode:         config.ModeBundle,
+			AbsOutputDir: "/out",
+		},
+	})
+}
+
+func TestTSExperimentalDecoratorsManglePropsDefineSemantics(t *testing.T) {
+	ts_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.ts": `
+				class Foo {
+					@dec(1) prop1 = null
+					@dec(2) prop2_ = null
+					@dec(3) ['prop3'] = null
+					@dec(4) ['prop4_'] = null
+					@dec(5) [/* @__KEY__ */ 'prop5'] = null
+					@dec(6) [/* @__KEY__ */ 'prop6_'] = null
+				}
+			`,
+			"/tsconfig.json": `{
+        "compilerOptions": {
+          "experimentalDecorators": true,
+          "useDefineForClassFields": true,
+        },
+			}`,
+		},
+		entryPaths: []string{"/entry.ts"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.js",
+			MangleProps:   regexp.MustCompile("_$"),
+		},
+	})
+}
+
+func TestTSExperimentalDecoratorsManglePropsAssignSemantics(t *testing.T) {
+	ts_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.ts": `
+				class Foo {
+					@dec(1) prop1 = null
+					@dec(2) prop2_ = null
+					@dec(3) ['prop3'] = null
+					@dec(4) ['prop4_'] = null
+					@dec(5) [/* @__KEY__ */ 'prop5'] = null
+					@dec(6) [/* @__KEY__ */ 'prop6_'] = null
+				}
+			`,
+			"/tsconfig.json": `{
+        "compilerOptions": {
+          "experimentalDecorators": true,
+          "useDefineForClassFields": false,
+        },
+			}`,
+		},
+		entryPaths: []string{"/entry.ts"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.js",
+			MangleProps:   regexp.MustCompile("_$"),
+		},
+	})
+}
+
+func TestTSExperimentalDecoratorsManglePropsMethods(t *testing.T) {
+	ts_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.ts": `
+				class Foo {
+					@dec(1) prop1() {}
+					@dec(2) prop2_() {}
+					@dec(3) ['prop3']() {}
+					@dec(4) ['prop4_']() {}
+					@dec(5) [/* @__KEY__ */ 'prop5']() {}
+					@dec(6) [/* @__KEY__ */ 'prop6_']() {}
+				}
+			`,
+			"/tsconfig.json": `{
+        "compilerOptions": {
+          "experimentalDecorators": true,
+        },
+			}`,
+		},
+		entryPaths: []string{"/entry.ts"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.js",
+			MangleProps:   regexp.MustCompile("_$"),
+		},
+	})
+}
+
+func TestTSExperimentalDecoratorsManglePropsStaticDefineSemantics(t *testing.T) {
+	ts_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.ts": `
+				class Foo {
+					@dec(1) static prop1 = null
+					@dec(2) static prop2_ = null
+					@dec(3) static ['prop3'] = null
+					@dec(4) static ['prop4_'] = null
+					@dec(5) static [/* @__KEY__ */ 'prop5'] = null
+					@dec(6) static [/* @__KEY__ */ 'prop6_'] = null
+				}
+			`,
+			"/tsconfig.json": `{
+        "compilerOptions": {
+          "experimentalDecorators": true,
+          "useDefineForClassFields": true,
+        },
+			}`,
+		},
+		entryPaths: []string{"/entry.ts"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.js",
+			MangleProps:   regexp.MustCompile("_$"),
+		},
+	})
+}
+
+func TestTSExperimentalDecoratorsManglePropsStaticAssignSemantics(t *testing.T) {
+	ts_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.ts": `
+				class Foo {
+					@dec(1) static prop1 = null
+					@dec(2) static prop2_ = null
+					@dec(3) static ['prop3'] = null
+					@dec(4) static ['prop4_'] = null
+					@dec(5) static [/* @__KEY__ */ 'prop5'] = null
+					@dec(6) static [/* @__KEY__ */ 'prop6_'] = null
+				}
+			`,
+			"/tsconfig.json": `{
+        "compilerOptions": {
+          "experimentalDecorators": true,
+          "useDefineForClassFields": false,
+        },
+			}`,
+		},
+		entryPaths: []string{"/entry.ts"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.js",
+			MangleProps:   regexp.MustCompile("_$"),
+		},
+	})
+}
+
+func TestTSExperimentalDecoratorsManglePropsStaticMethods(t *testing.T) {
+	ts_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.ts": `
+				class Foo {
+					@dec(1) static prop1() {}
+					@dec(2) static prop2_() {}
+					@dec(3) static ['prop3']() {}
+					@dec(4) static ['prop4_']() {}
+					@dec(5) static [/* @__KEY__ */ 'prop5']() {}
+					@dec(6) static [/* @__KEY__ */ 'prop6_']() {}
+				}
+			`,
+			"/tsconfig.json": `{
+        "compilerOptions": {
+          "experimentalDecorators": true,
+        },
+			}`,
+		},
+		entryPaths: []string{"/entry.ts"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.js",
+			MangleProps:   regexp.MustCompile("_$"),
+		},
+	})
+}
+
+func TestTSPrintNonFiniteNumberInsideWith(t *testing.T) {
+	ts_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.ts": `
+				// Use const enums to force inline values
+				const enum Foo {
+					NAN = 0 / 0,
+					POS_INF = 1 / 0,
+					NEG_INF = -1 / 0,
+				}
+
+				//! It's ok to use "NaN" and "Infinity" here
+				console.log(
+					Foo.NAN,
+					Foo.POS_INF,
+					Foo.NEG_INF,
+				)
+				checkPrecedence(
+					1 / Foo.NAN,
+					1 / Foo.POS_INF,
+					1 / Foo.NEG_INF,
+				)
+
+				//! We must not use "NaN" or "Infinity" inside "with"
+				with (x) {
+					console.log(
+						Foo.NAN,
+						Foo.POS_INF,
+						Foo.NEG_INF,
+					)
+					checkPrecedence(
+						1 / Foo.NAN,
+						1 / Foo.POS_INF,
+						1 / Foo.NEG_INF,
+					)
+				}
+			`,
+		},
+		entryPaths: []string{"/entry.ts"},
+		options: config.Options{
+			Mode:          config.ModePassThrough,
+			AbsOutputFile: "/out.js",
+		},
+	})
+}
+
+func TestTSImportInNodeModulesNameCollisionWithCSS(t *testing.T) {
+	ts_suite.expectBundled(t, bundled{
+		files: map[string]string{
+			"/entry.ts": `
+				import "pkg"
+			`,
+			"/node_modules/pkg/index.ts": `
+				import js_ts_css from "./js_ts_css"
+				import ts_css from "./ts_css"
+				import js_ts from "./js_ts"
+				js_ts_css()
+				ts_css()
+				js_ts()
+			`,
+			"/node_modules/pkg/js_ts_css.js": `
+				import './js_ts_css.css'
+				export default function() {}
+			`,
+			"/node_modules/pkg/js_ts_css.ts": `
+				TEST FAILED
+			`,
+			"/node_modules/pkg/js_ts_css.css": `
+				.js_ts_css {}
+			`,
+			"/node_modules/pkg/ts_css.ts": `
+				import './ts_css.css'
+				export default function() {}
+			`,
+			"/node_modules/pkg/ts_css.css": `
+				.ts_css {}
+			`,
+			"/node_modules/pkg/js_ts.js": `
+				export default function() {}
+			`,
+			"/node_modules/pkg/js_ts.ts": `
+				TEST FAILED
+			`,
+		},
+		entryPaths: []string{"/entry.ts"},
+		options: config.Options{
+			Mode:          config.ModeBundle,
+			AbsOutputFile: "/out.js",
+		},
+	})
+}
