@@ -13,13 +13,6 @@ import (
 	"github.com/evanw/esbuild/internal/test"
 )
 
-func assertEqual(t *testing.T, a interface{}, b interface{}) {
-	t.Helper()
-	if a != b {
-		t.Fatalf("%s != %s", a, b)
-	}
-}
-
 func expectPrintedCommon(t *testing.T, name string, contents string, expected string, options config.Options) {
 	t.Helper()
 	t.Run(name, func(t *testing.T) {
@@ -511,6 +504,17 @@ func TestObject(t *testing.T) {
 	expectPrinted(t, "let x = () => ({}.x)", "let x = () => ({}).x;\n")
 	expectPrinted(t, "let x = () => ({} = {})", "let x = () => ({} = {});\n")
 	expectPrinted(t, "let x = () => (x, {} = {})", "let x = () => (x, {} = {});\n")
+
+	// "{ __proto__: __proto__ }" must not become "{ __proto__ }"
+	expectPrinted(t, "function foo(__proto__) { return { __proto__: __proto__ } }", "function foo(__proto__) {\n  return { __proto__: __proto__ };\n}\n")
+	expectPrinted(t, "function foo(__proto__) { return { '__proto__': __proto__ } }", "function foo(__proto__) {\n  return { \"__proto__\": __proto__ };\n}\n")
+	expectPrinted(t, "function foo(__proto__) { return { ['__proto__']: __proto__ } }", "function foo(__proto__) {\n  return { [\"__proto__\"]: __proto__ };\n}\n")
+	expectPrinted(t, "import { __proto__ } from 'foo'; let foo = () => ({ __proto__: __proto__ })", "import { __proto__ } from \"foo\";\nlet foo = () => ({ __proto__: __proto__ });\n")
+	expectPrinted(t, "import { __proto__ } from 'foo'; let foo = () => ({ '__proto__': __proto__ })", "import { __proto__ } from \"foo\";\nlet foo = () => ({ \"__proto__\": __proto__ });\n")
+	expectPrinted(t, "import { __proto__ } from 'foo'; let foo = () => ({ ['__proto__']: __proto__ })", "import { __proto__ } from \"foo\";\nlet foo = () => ({ [\"__proto__\"]: __proto__ });\n")
+
+	// Don't use ES6+ features (such as a shorthand or computed property name) in ES5
+	expectPrintedTarget(t, 5, "function foo(__proto__) { return { __proto__ } }", "function foo(__proto__) {\n  return { __proto__: __proto__ };\n}\n")
 }
 
 func TestFor(t *testing.T) {
@@ -707,6 +711,13 @@ func TestPrivateIdentifiers(t *testing.T) {
 	expectPrintedMinify(t, "class Foo { #foo; foo() { return #foo in this } }", "class Foo{#foo;foo(){return#foo in this}}")
 }
 
+func TestDecorators(t *testing.T) {
+	example := "class Foo {\n@w\nw; @x x; @a1\n@b1@b2\n@c1@c2@c3\ny = @y1 @y2 class {}; @a1\n@b1@b2\n@c1@c2@c3 z =\n@z1\n@z2\nclass {}}"
+	expectPrinted(t, example, "class Foo {\n  @w\n  w;\n  @x x;\n  @a1\n  @b1 @b2\n  @c1 @c2 @c3\n  "+
+		"y = @y1 @y2 class {\n  };\n  @a1\n  @b1 @b2\n  @c1 @c2 @c3 z = @z1 @z2 class {\n  };\n}\n")
+	expectPrintedMinify(t, example, "class Foo{@w w;@x x;@a1@b1@b2@c1@c2@c3 y=@y1@y2 class{};@a1@b1@b2@c1@c2@c3 z=@z1@z2 class{}}")
+}
+
 func TestImport(t *testing.T) {
 	expectPrinted(t, "import('path');", "import(\"path\");\n") // The semicolon must not be a separate statement
 
@@ -805,7 +816,7 @@ func TestWhitespace(t *testing.T) {
 
 	expectPrintedMinify(t, "()=>({})", "()=>({});")
 	expectPrintedMinify(t, "()=>({}[1])", "()=>({})[1];")
-	expectPrintedMinify(t, "()=>({}+0)", "()=>({}+0);")
+	expectPrintedMinify(t, "()=>({}+0)", "()=>\"[object Object]0\";")
 	expectPrintedMinify(t, "()=>function(){}", "()=>function(){};")
 
 	expectPrintedMinify(t, "(function(){})", "(function(){});")
@@ -939,31 +950,41 @@ func TestJSX(t *testing.T) {
 	expectPrintedJSX(t, "<a b></a>", "<a b />;\n")
 
 	expectPrintedJSX(t, "<a b={true}></a>", "<a b={true} />;\n")
-	expectPrintedJSX(t, "<a b={'x'}></a>", "<a b=\"x\" />;\n")
-	expectPrintedJSX(t, "<a b={`'`}></a>", "<a b=\"'\" />;\n")
-	expectPrintedJSX(t, "<a b={`\"`}></a>", "<a b='\"' />;\n")
+	expectPrintedJSX(t, "<a b='x'></a>", "<a b='x' />;\n")
+	expectPrintedJSX(t, "<a b=\"x\"></a>", "<a b=\"x\" />;\n")
+	expectPrintedJSX(t, "<a b={'x'}></a>", "<a b={\"x\"} />;\n")
+	expectPrintedJSX(t, "<a b={`'`}></a>", "<a b={`'`} />;\n")
+	expectPrintedJSX(t, "<a b={`\"`}></a>", "<a b={`\"`} />;\n")
 	expectPrintedJSX(t, "<a b={`'\"`}></a>", "<a b={`'\"`} />;\n")
-	expectPrintedJSX(t, "<a b=\"&quot;\"></a>", "<a b='\"' />;\n")
-	expectPrintedJSX(t, "<a b=\"&amp;\"></a>", "<a b={\"&\"} />;\n")
+	expectPrintedJSX(t, "<a b=\"&quot;\"></a>", "<a b=\"&quot;\" />;\n")
+	expectPrintedJSX(t, "<a b=\"&amp;\"></a>", "<a b=\"&amp;\" />;\n")
 
 	expectPrintedJSX(t, "<a>x</a>", "<a>x</a>;\n")
-	expectPrintedJSX(t, "<a>x\ny</a>", "<a>x y</a>;\n")
-	expectPrintedJSX(t, "<a>{'x'}{'y'}</a>", "<a>\n  {\"x\"}\n  {\"y\"}\n</a>;\n")
+	expectPrintedJSX(t, "<a>x\ny</a>", "<a>x\ny</a>;\n")
+	expectPrintedJSX(t, "<a>{'x'}{'y'}</a>", "<a>{\"x\"}{\"y\"}</a>;\n")
 	expectPrintedJSX(t, "<a> x</a>", "<a> x</a>;\n")
 	expectPrintedJSX(t, "<a>x </a>", "<a>x </a>;\n")
-	expectPrintedJSX(t, "<a>&#10;</a>", "<a>{\"\\n\"}</a>;\n")
-	expectPrintedJSX(t, "<a>&amp;</a>", "<a>{\"&\"}</a>;\n")
-	expectPrintedJSX(t, "<a>&lt;</a>", "<a>{\"<\"}</a>;\n")
-	expectPrintedJSX(t, "<a>&gt;</a>", "<a>{\">\"}</a>;\n")
-	expectPrintedJSX(t, "<a>&#123;</a>", "<a>{\"{\"}</a>;\n")
-	expectPrintedJSX(t, "<a>&#125;</a>", "<a>{\"}\"}</a>;\n")
+	expectPrintedJSX(t, "<a>&#10;</a>", "<a>&#10;</a>;\n")
+	expectPrintedJSX(t, "<a>&amp;</a>", "<a>&amp;</a>;\n")
+	expectPrintedJSX(t, "<a>&lt;</a>", "<a>&lt;</a>;\n")
+	expectPrintedJSX(t, "<a>&gt;</a>", "<a>&gt;</a>;\n")
+	expectPrintedJSX(t, "<a>&#123;</a>", "<a>&#123;</a>;\n")
+	expectPrintedJSX(t, "<a>&#125;</a>", "<a>&#125;</a>;\n")
 
 	expectPrintedJSX(t, "<a><x/></a>", "<a><x /></a>;\n")
-	expectPrintedJSX(t, "<a><x/><y/></a>", "<a>\n  <x />\n  <y />\n</a>;\n")
-	expectPrintedJSX(t, "<a>b<c/>d</a>", "<a>\n  {\"b\"}\n  <c />\n  {\"d\"}\n</a>;\n")
+	expectPrintedJSX(t, "<a><x/><y/></a>", "<a><x /><y /></a>;\n")
+	expectPrintedJSX(t, "<a>b<c/>d</a>", "<a>b<c />d</a>;\n")
 
 	expectPrintedJSX(t, "<></>", "<></>;\n")
-	expectPrintedJSX(t, "<>x<y/>z</>", "<>\n  {\"x\"}\n  <y />\n  {\"z\"}\n</>;\n")
+	expectPrintedJSX(t, "<>x<y/>z</>", "<>x<y />z</>;\n")
+
+	// JSX elements as JSX attribute values
+	expectPrintedJSX(t, "<a b=<c/>/>", "<a b=<c /> />;\n")
+	expectPrintedJSX(t, "<a b=<>c</>/>", "<a b=<>c</> />;\n")
+	expectPrintedJSX(t, "<a b=<>{c}</>/>", "<a b=<>{c}</> />;\n")
+	expectPrintedJSX(t, "<a b={<c/>}/>", "<a b={<c />} />;\n")
+	expectPrintedJSX(t, "<a b={<>c</>}/>", "<a b={<>c</>} />;\n")
+	expectPrintedJSX(t, "<a b={<>{c}</>}/>", "<a b={<>{c}</>} />;\n")
 
 	// These can't be escaped because JSX lacks a syntax for escapes
 	expectPrintedJSXASCII(t, "<π/>", "<π />;\n")
@@ -974,16 +995,23 @@ func TestJSX(t *testing.T) {
 	expectPrintedJSXASCII(t, "<a π/>", "<a π />;\n")
 	expectPrintedJSXASCII(t, "<a 𐀀/>", "<a 𐀀 />;\n")
 
-	// These can be escaped but as JS strings (since XML entities in JSX aren't standard)
-	expectPrintedJSXASCII(t, "<a b='π'/>", "<a b={\"\\u03C0\"} />;\n")
-	expectPrintedJSXASCII(t, "<a b='𐀀'/>", "<a b={\"\\u{10000}\"} />;\n")
-	expectPrintedJSXASCII(t, "<a>π</a>", "<a>{\"\\u03C0\"}</a>;\n")
-	expectPrintedJSXASCII(t, "<a>𐀀</a>", "<a>{\"\\u{10000}\"}</a>;\n")
+	// JSX text is deliberately not printed as ASCII when JSX preservation is
+	// enabled. This is because:
+	//
+	// a) The JSX specification doesn't say how JSX text is supposed to be interpreted
+	// b) Enabling JSX preservation means that JSX will be transformed again anyway
+	// c) People do very weird/custom things with JSX that "preserve" shouldn't break
+	//
+	// See also: https://github.com/evanw/esbuild/issues/3605
+	expectPrintedJSXASCII(t, "<a b='π'/>", "<a b='π' />;\n")
+	expectPrintedJSXASCII(t, "<a b='𐀀'/>", "<a b='𐀀' />;\n")
+	expectPrintedJSXASCII(t, "<a>π</a>", "<a>π</a>;\n")
+	expectPrintedJSXASCII(t, "<a>𐀀</a>", "<a>𐀀</a>;\n")
 
-	expectPrintedJSXMinify(t, "<a b c={x,y} d='true'/>", "<a b c={(x,y)}d=\"true\"/>;")
+	expectPrintedJSXMinify(t, "<a b c={x,y} d='true'/>", "<a b c={(x,y)}d='true'/>;")
 	expectPrintedJSXMinify(t, "<a><b/><c/></a>", "<a><b/><c/></a>;")
 	expectPrintedJSXMinify(t, "<a> x <b/> y </a>", "<a> x <b/> y </a>;")
-	expectPrintedJSXMinify(t, "<a>{' x '}{'<b/>'}{' y '}</a>", "<a> x {\"<b/>\"} y </a>;")
+	expectPrintedJSXMinify(t, "<a>{' x '}{'<b/>'}{' y '}</a>", "<a>{\" x \"}{\"<b/>\"}{\" y \"}</a>;")
 }
 
 func TestJSXSingleLine(t *testing.T) {
