@@ -21,6 +21,14 @@ type packageJSON struct {
 	mainFields     map[string]mainField
 	moduleTypeData js_ast.ModuleTypeData
 
+	// "TypeScript will first check whether package.json contains a "tsconfig"
+	// field, and if it does, TypeScript will try to load a configuration file
+	// from that field. If neither exists, TypeScript will try to read from a
+	// tsconfig.json at the root."
+	//
+	// See: https://www.typescriptlang.org/docs/handbook/release-notes/typescript-3-2.html#tsconfigjson-inheritance-via-nodejs-packages
+	tsconfig string
+
 	// Present if the "browser" field is present. This field is intended to be
 	// used by bundlers and lets you redirect the paths of certain 3rd-party
 	// modules that don't work in the browser to other modules that shim that
@@ -323,6 +331,13 @@ func (r resolverQuery) parsePackageJSON(inputPath string) *packageJSON {
 		}
 	}
 
+	// Read the "tsconfig" field
+	if tsconfigJSON, _, ok := getProperty(json, "tsconfig"); ok {
+		if tsconfigValue, ok := getString(tsconfigJSON); ok {
+			packageJSON.tsconfig = tsconfigValue
+		}
+	}
+
 	// Read the "main" fields
 	mainFields := r.options.MainFields
 	if mainFields == nil {
@@ -443,7 +458,7 @@ func (r resolverQuery) parsePackageJSON(inputPath string) *packageJSON {
 
 	// Read the "imports" map
 	if importsJSON, importsLoc, ok := getProperty(json, "imports"); ok {
-		if importsMap := parseImportsExportsMap(jsonSource, r.log, importsJSON, importsLoc); importsMap != nil {
+		if importsMap := parseImportsExportsMap(jsonSource, r.log, importsJSON, "imports", importsLoc); importsMap != nil {
 			if importsMap.root.kind != pjObject {
 				r.log.AddID(logger.MsgID_PackageJSON_InvalidImportsOrExports, logger.Warning, &tracker, importsMap.root.firstToken,
 					"The value for \"imports\" must be an object")
@@ -454,7 +469,7 @@ func (r resolverQuery) parsePackageJSON(inputPath string) *packageJSON {
 
 	// Read the "exports" map
 	if exportsJSON, exportsLoc, ok := getProperty(json, "exports"); ok {
-		if exportsMap := parseImportsExportsMap(jsonSource, r.log, exportsJSON, exportsLoc); exportsMap != nil {
+		if exportsMap := parseImportsExportsMap(jsonSource, r.log, exportsJSON, "exports", exportsLoc); exportsMap != nil {
 			packageJSON.exportsMap = exportsMap
 		}
 	}
@@ -525,6 +540,7 @@ func globstarToEscapedRegexp(glob string) (string, bool) {
 // Reference: https://nodejs.org/api/esm.html#esm_resolver_algorithm_specification
 type pjMap struct {
 	root           pjEntry
+	propertyKey    string
 	propertyKeyLoc logger.Loc
 }
 
@@ -621,7 +637,7 @@ func (entry pjEntry) valueForKey(key string) (pjEntry, bool) {
 	return pjEntry{}, false
 }
 
-func parseImportsExportsMap(source logger.Source, log logger.Log, json js_ast.Expr, propertyKeyLoc logger.Loc) *pjMap {
+func parseImportsExportsMap(source logger.Source, log logger.Log, json js_ast.Expr, propertyKey string, propertyKeyLoc logger.Loc) *pjMap {
 	var visit func(expr js_ast.Expr) pjEntry
 	tracker := logger.MakeLineColumnTracker(&source)
 
@@ -730,7 +746,11 @@ func parseImportsExportsMap(source logger.Source, log logger.Log, json js_ast.Ex
 		return nil
 	}
 
-	return &pjMap{root: root, propertyKeyLoc: propertyKeyLoc}
+	return &pjMap{
+		root:           root,
+		propertyKey:    propertyKey,
+		propertyKeyLoc: propertyKeyLoc,
+	}
 }
 
 func (entry pjEntry) keysStartWithDot() bool {
@@ -788,6 +808,9 @@ type pjDebug struct {
 
 	// This is the range of the token to use for error messages
 	token logger.Range
+
+	// If true, the token is a "null" literal
+	isBecauseOfNullLiteral bool
 }
 
 func (r resolverQuery) esmHandlePostConditions(
@@ -872,6 +895,7 @@ func (r resolverQuery) esmPackageExportsResolve(
 		return "", pjStatusInvalidPackageConfiguration, pjDebug{token: exports.firstToken}
 	}
 
+	debugToReturn := pjDebug{token: exports.firstToken}
 	if subpath == "." {
 		mainExport := pjEntry{kind: pjNull}
 		if exports.kind == pjString || exports.kind == pjArray || (exports.kind == pjObject && !exports.keysStartWithDot()) {
@@ -888,19 +912,23 @@ func (r resolverQuery) esmPackageExportsResolve(
 			resolved, status, debug := r.esmPackageTargetResolve(packageURL, mainExport, "", false, false, conditions)
 			if status != pjStatusNull && status != pjStatusUndefined {
 				return resolved, status, debug
+			} else {
+				debugToReturn = debug
 			}
 		}
 	} else if exports.kind == pjObject && exports.keysStartWithDot() {
 		resolved, status, debug := r.esmPackageImportsExportsResolve(subpath, exports, packageURL, false, conditions)
 		if status != pjStatusNull && status != pjStatusUndefined {
 			return resolved, status, debug
+		} else {
+			debugToReturn = debug
 		}
 	}
 
 	if r.debugLogs != nil {
 		r.debugLogs.addNote(fmt.Sprintf("The path %q is not exported", subpath))
 	}
-	return "", pjStatusPackagePathNotExported, pjDebug{token: exports.firstToken}
+	return "", pjStatusPackagePathNotExported, debugToReturn
 }
 
 func (r resolverQuery) esmPackageImportsExportsResolve(
@@ -1155,14 +1183,14 @@ func (r resolverQuery) esmPackageTargetResolve(
 				//
 				// We want the warning to say this:
 				//
-				//   note: None of the conditions provided ("require") match any of the
+				//   note: None of the conditions in the package definition ("require") match any of the
 				//         currently active conditions ("default", "import", "node")
 				//   14 |       "node": {
 				//      |               ^
 				//
 				// We don't want the warning to say this:
 				//
-				//   note: None of the conditions provided ("browser", "electron", "node")
+				//   note: None of the conditions in the package definition ("browser", "electron", "node")
 				//         match any of the currently active conditions ("default", "import", "node")
 				//   7 |   "exports": {
 				//     |              ^
@@ -1217,7 +1245,7 @@ func (r resolverQuery) esmPackageTargetResolve(
 		if r.debugLogs != nil {
 			r.debugLogs.addNote(fmt.Sprintf("The path %q is set to null", subpath))
 		}
-		return "", pjStatusNull, pjDebug{token: target.firstToken}
+		return "", pjStatusNull, pjDebug{token: target.firstToken, isBecauseOfNullLiteral: true}
 	}
 
 	if r.debugLogs != nil {
@@ -1348,7 +1376,6 @@ func (r resolverQuery) esmPackageTargetReverseResolve(
 					return true, keyWithoutTrailingStar + starData, target.firstToken
 				}
 			}
-			break
 		}
 
 	case pjObject:
