@@ -145,7 +145,9 @@ function pushCommonFlags(flags: string[], options: CommonOptions, keys: OptionKe
   let minifySyntax = getFlag(options, keys, 'minifySyntax', mustBeBoolean)
   let minifyWhitespace = getFlag(options, keys, 'minifyWhitespace', mustBeBoolean)
   let minifyIdentifiers = getFlag(options, keys, 'minifyIdentifiers', mustBeBoolean)
+  let lineLimit = getFlag(options, keys, 'lineLimit', mustBeInteger)
   let drop = getFlag(options, keys, 'drop', mustBeArray)
+  let dropLabels = getFlag(options, keys, 'dropLabels', mustBeArray)
   let charset = getFlag(options, keys, 'charset', mustBeString)
   let treeShaking = getFlag(options, keys, 'treeShaking', mustBeBoolean)
   let ignoreAnnotations = getFlag(options, keys, 'ignoreAnnotations', mustBeBoolean)
@@ -161,6 +163,7 @@ function pushCommonFlags(flags: string[], options: CommonOptions, keys: OptionKe
   let pure = getFlag(options, keys, 'pure', mustBeArray)
   let keepNames = getFlag(options, keys, 'keepNames', mustBeBoolean)
   let platform = getFlag(options, keys, 'platform', mustBeString)
+  let tsconfigRaw = getFlag(options, keys, 'tsconfigRaw', mustBeStringOrObject)
 
   if (legalComments) flags.push(`--legal-comments=${legalComments}`)
   if (sourceRoot !== void 0) flags.push(`--source-root=${sourceRoot}`)
@@ -172,15 +175,18 @@ function pushCommonFlags(flags: string[], options: CommonOptions, keys: OptionKe
   if (format) flags.push(`--format=${format}`)
   if (globalName) flags.push(`--global-name=${globalName}`)
   if (platform) flags.push(`--platform=${platform}`)
+  if (tsconfigRaw) flags.push(`--tsconfig-raw=${typeof tsconfigRaw === 'string' ? tsconfigRaw : JSON.stringify(tsconfigRaw)}`)
 
   if (minify) flags.push('--minify')
   if (minifySyntax) flags.push('--minify-syntax')
   if (minifyWhitespace) flags.push('--minify-whitespace')
   if (minifyIdentifiers) flags.push('--minify-identifiers')
+  if (lineLimit) flags.push(`--line-limit=${lineLimit}`)
   if (charset) flags.push(`--charset=${charset}`)
   if (treeShaking !== void 0) flags.push(`--tree-shaking=${treeShaking}`)
   if (ignoreAnnotations) flags.push(`--ignore-annotations`)
   if (drop) for (let what of drop) flags.push(`--drop:${validateStringValue(what, 'drop')}`)
+  if (dropLabels) flags.push(`--drop-labels=${Array.from(dropLabels).map(what => validateStringValue(what, 'dropLabels')).join(',')}`)
   if (mangleProps) flags.push(`--mangle-props=${mangleProps.source}`)
   if (reserveProps) flags.push(`--reserve-props=${reserveProps.source}`)
   if (mangleQuoted !== void 0) flags.push(`--mangle-quoted=${mangleQuoted}`)
@@ -422,7 +428,6 @@ function flagsForTransformOptions(
   pushCommonFlags(flags, options, keys)
 
   let sourcemap = getFlag(options, keys, 'sourcemap', mustBeStringOrBoolean)
-  let tsconfigRaw = getFlag(options, keys, 'tsconfigRaw', mustBeStringOrObject)
   let sourcefile = getFlag(options, keys, 'sourcefile', mustBeString)
   let loader = getFlag(options, keys, 'loader', mustBeString)
   let banner = getFlag(options, keys, 'banner', mustBeString)
@@ -431,7 +436,6 @@ function flagsForTransformOptions(
   checkForInvalidFlags(options, keys, `in ${callName}() call`)
 
   if (sourcemap) flags.push(`--sourcemap=${sourcemap === true ? 'external' : sourcemap}`)
-  if (tsconfigRaw) flags.push(`--tsconfig-raw=${typeof tsconfigRaw === 'string' ? tsconfigRaw : JSON.stringify(tsconfigRaw)}`)
   if (sourcefile) flags.push(`--sourcefile=${sourcefile}`)
   if (loader) flags.push(`--loader=${loader}`)
   if (banner) flags.push(`--banner=${banner}`)
@@ -588,18 +592,31 @@ export function createChannel(streamIn: StreamIn): StreamOut {
 
       if (typeof request.key === 'number') {
         const requestCallbacks = requestCallbacksByKey[request.key]
-        if (requestCallbacks) {
-          const callback = requestCallbacks[request.command]
-          if (callback) {
-            await callback(id, request)
-            return
-          }
+        if (!requestCallbacks) {
+          // Ignore invalid commands for old builds that no longer exist.
+          // This can happen when "context.cancel" and "context.dispose"
+          // is called while esbuild is processing many files in parallel.
+          // See https://github.com/evanw/esbuild/issues/3318 for details.
+          return
+        }
+        const callback = requestCallbacks[request.command]
+        if (callback) {
+          await callback(id, request)
+          return
         }
       }
 
       throw new Error(`Invalid command: ` + request.command)
     } catch (e) {
-      sendResponse(id, { errors: [extractErrorMessageV8(e, streamIn, null, void 0, '')] } as any)
+      const errors = [extractErrorMessageV8(e, streamIn, null, void 0, '')]
+      try {
+        sendResponse(id, { errors } as any)
+      } catch {
+        // This may fail if the esbuild process is no longer running, but
+        // that's ok. Catch and swallow this exception so that we don't
+        // cause an unhandled promise rejection. Our caller isn't expecting
+        // this call to fail and doesn't handle the promise rejection.
+      }
     }
   }
 
@@ -783,7 +800,6 @@ export function createChannel(streamIn: StreamIn): StreamOut {
   }
 
   let formatMessages: StreamService['formatMessages'] = ({ callName, refs, messages, options, callback }) => {
-    let result = sanitizeMessages(messages, 'messages', null, '')
     if (!options) throw new Error(`Missing second argument in ${callName}() call`)
     let keys: OptionKeys = {}
     let kind = getFlag(options, keys, 'kind', mustBeString)
@@ -794,7 +810,7 @@ export function createChannel(streamIn: StreamIn): StreamOut {
     if (kind !== 'error' && kind !== 'warning') throw new Error(`Expected "kind" to be "error" or "warning" in ${callName}() call`)
     let request: protocol.FormatMsgsRequest = {
       command: 'format-msgs',
-      messages: result,
+      messages: sanitizeMessages(messages, 'messages', null, '', terminalWidth),
       isWarning: kind === 'warning',
     }
     if (color !== void 0) request.color = color
@@ -1080,6 +1096,7 @@ function buildOrContextImpl(
           const servedir = getFlag(options, keys, 'servedir', mustBeString)
           const keyfile = getFlag(options, keys, 'keyfile', mustBeString)
           const certfile = getFlag(options, keys, 'certfile', mustBeString)
+          const fallback = getFlag(options, keys, 'fallback', mustBeString)
           const onRequest = getFlag(options, keys, 'onRequest', mustBeFunction)
           checkForInvalidFlags(options, keys, `in serve() call`)
 
@@ -1093,6 +1110,7 @@ function buildOrContextImpl(
           if (servedir !== void 0) request.servedir = servedir
           if (keyfile !== void 0) request.keyfile = keyfile
           if (certfile !== void 0) request.certfile = certfile
+          if (fallback !== void 0) request.fallback = fallback
 
           sendRequest<protocol.ServeRequest, protocol.ServeResponse>(refs, request, (error, response) => {
             if (error) return reject(new Error(error))
@@ -1342,8 +1360,8 @@ let handlePlugins = async (
           let warnings = getFlag(result, keys, 'warnings', mustBeArray)
           checkForInvalidFlags(result, keys, `from onStart() callback in plugin ${quote(name)}`)
 
-          if (errors != null) response.errors!.push(...sanitizeMessages(errors, 'errors', details, name))
-          if (warnings != null) response.warnings!.push(...sanitizeMessages(warnings, 'warnings', details, name))
+          if (errors != null) response.errors!.push(...sanitizeMessages(errors, 'errors', details, name, undefined))
+          if (warnings != null) response.warnings!.push(...sanitizeMessages(warnings, 'warnings', details, name, undefined))
         }
       } catch (e) {
         response.errors!.push(extractErrorMessageV8(e, streamIn, details, note && note(), name))
@@ -1390,8 +1408,8 @@ let handlePlugins = async (
           if (external != null) response.external = external
           if (sideEffects != null) response.sideEffects = sideEffects
           if (pluginData != null) response.pluginData = details.store(pluginData)
-          if (errors != null) response.errors = sanitizeMessages(errors, 'errors', details, name)
-          if (warnings != null) response.warnings = sanitizeMessages(warnings, 'warnings', details, name)
+          if (errors != null) response.errors = sanitizeMessages(errors, 'errors', details, name, undefined)
+          if (warnings != null) response.warnings = sanitizeMessages(warnings, 'warnings', details, name, undefined)
           if (watchFiles != null) response.watchFiles = sanitizeStringArray(watchFiles, 'watchFiles')
           if (watchDirs != null) response.watchDirs = sanitizeStringArray(watchDirs, 'watchDirs')
           break
@@ -1414,6 +1432,7 @@ let handlePlugins = async (
           namespace: request.namespace,
           suffix: request.suffix,
           pluginData: details.load(request.pluginData),
+          with: request.with,
         })
 
         if (result != null) {
@@ -1437,8 +1456,8 @@ let handlePlugins = async (
           if (resolveDir != null) response.resolveDir = resolveDir
           if (pluginData != null) response.pluginData = details.store(pluginData)
           if (loader != null) response.loader = loader
-          if (errors != null) response.errors = sanitizeMessages(errors, 'errors', details, name)
-          if (warnings != null) response.warnings = sanitizeMessages(warnings, 'warnings', details, name)
+          if (errors != null) response.errors = sanitizeMessages(errors, 'errors', details, name, undefined)
+          if (warnings != null) response.warnings = sanitizeMessages(warnings, 'warnings', details, name, undefined)
           if (watchFiles != null) response.watchFiles = sanitizeStringArray(watchFiles, 'watchFiles')
           if (watchDirs != null) response.watchDirs = sanitizeStringArray(watchDirs, 'watchDirs')
           break
@@ -1473,8 +1492,8 @@ let handlePlugins = async (
               let warnings = getFlag(value, keys, 'warnings', mustBeArray)
               checkForInvalidFlags(value, keys, `from onEnd() callback in plugin ${quote(name)}`)
 
-              if (errors != null) newErrors = sanitizeMessages(errors, 'errors', details, name)
-              if (warnings != null) newWarnings = sanitizeMessages(warnings, 'warnings', details, name)
+              if (errors != null) newErrors = sanitizeMessages(errors, 'errors', details, name, undefined)
+              if (warnings != null) newWarnings = sanitizeMessages(warnings, 'warnings', details, name, undefined)
             }
           } catch (e) {
             newErrors = [extractErrorMessageV8(e, streamIn, details, note && note(), name)]
@@ -1639,7 +1658,7 @@ function parseStackLinesV8(streamIn: StreamIn, lines: string[], ident: string):
 
 function failureErrorWithLog(text: string, errors: types.Message[], warnings: types.Message[]): types.BuildFailure {
   let limit = 5
-  let summary = errors.length < 1 ? '' : ` with ${errors.length} error${errors.length < 2 ? '' : 's'}:` +
+  text += errors.length < 1 ? '' : ` with ${errors.length} error${errors.length < 2 ? '' : 's'}:` +
     errors.slice(0, limit + 1).map((e, i) => {
       if (i === limit) return '\n...'
       if (!e.location) return `\nerror: ${e.text}`
@@ -1647,9 +1666,26 @@ function failureErrorWithLog(text: string, errors: types.Message[], warnings: ty
       let pluginText = e.pluginName ? `[plugin: ${e.pluginName}] ` : ''
       return `\n${file}:${line}:${column}: ERROR: ${pluginText}${e.text}`
     }).join('')
-  let error: any = new Error(`${text}${summary}`)
-  error.errors = errors
-  error.warnings = warnings
+  let error: any = new Error(text)
+
+  // Use a getter instead of a plain property so that when the error is thrown
+  // without being caught and the node process exits, the error objects aren't
+  // printed. The error objects are pretty big and not helpful because a) esbuild
+  // already prints errors to stderr by default and b) the error summary already
+  // has a more helpful abbreviated form of the error messages.
+  for (const [key, value] of [['errors', errors], ['warnings', warnings]] as const) {
+    Object.defineProperty(error, key, {
+      configurable: true,
+      enumerable: true,
+      get: () => value,
+      set: value => Object.defineProperty(error, key, {
+        configurable: true,
+        enumerable: true,
+        value,
+      }),
+    })
+  }
+
   return error
 }
 
@@ -1660,7 +1696,7 @@ function replaceDetailsInMessages(messages: types.Message[], stash: ObjectStash)
   return messages
 }
 
-function sanitizeLocation(location: types.PartialMessage['location'], where: string): types.Message['location'] {
+function sanitizeLocation(location: types.PartialMessage['location'], where: string, terminalWidth: number | undefined): types.Message['location'] {
   if (location == null) return null
 
   let keys: OptionKeys = {}
@@ -1673,6 +1709,32 @@ function sanitizeLocation(location: types.PartialMessage['location'], where: str
   let suggestion = getFlag(location, keys, 'suggestion', mustBeString)
   checkForInvalidFlags(location, keys, where)
 
+  // Performance hack: Some people pass enormous minified files as the line
+  // text with a column near the beginning of the line and then complain
+  // when this function is slow. The slowness comes from serializing a huge
+  // string. But the vast majority of that string is unnecessary. Try to
+  // detect when this is the case and trim the string before serialization
+  // to avoid the performance hit. See: https://github.com/evanw/esbuild/issues/3467
+  if (lineText) {
+    // Try to conservatively guess the maximum amount of relevant text
+    const relevantASCII = lineText.slice(0,
+      (column && column > 0 ? column : 0) +
+      (length && length > 0 ? length : 0) +
+      (terminalWidth && terminalWidth > 0 ? terminalWidth : 80))
+
+    // Make sure it's ASCII (so the byte-oriented column and length values
+    // are correct) and that there are no newlines (so that our logging code
+    // doesn't look at the end of the string)
+    if (!/[\x7F-\uFFFF]/.test(relevantASCII) && !/\n/.test(lineText)) {
+      lineText = relevantASCII
+    }
+  }
+
+  // Note: We could technically make this even faster by maintaining two copies
+  // of this code, one in Go and one in TypeScript. But I'm not going to do that.
+  // The point of this function is to call into the real Go code to get what it
+  // does. If someone wants a JS version, they can port it themselves.
+
   return {
     file: file || '',
     namespace: namespace || '',
@@ -1684,7 +1746,13 @@ function sanitizeLocation(location: types.PartialMessage['location'], where: str
   }
 }
 
-function sanitizeMessages(messages: types.PartialMessage[], property: string, stash: ObjectStash | null, fallbackPluginName: string): types.Message[] {
+function sanitizeMessages(
+  messages: types.PartialMessage[],
+  property: string,
+  stash: ObjectStash | null,
+  fallbackPluginName: string,
+  terminalWidth: number | undefined,
+): types.Message[] {
   let messagesClone: types.Message[] = []
   let index = 0
 
@@ -1708,7 +1776,7 @@ function sanitizeMessages(messages: types.PartialMessage[], property: string, st
         checkForInvalidFlags(note, noteKeys, where)
         notesClone.push({
           text: noteText || '',
-          location: sanitizeLocation(noteLocation, where),
+          location: sanitizeLocation(noteLocation, where, terminalWidth),
         })
       }
     }
@@ -1717,7 +1785,7 @@ function sanitizeMessages(messages: types.PartialMessage[], property: string, st
       id: id || '',
       pluginName: pluginName || fallbackPluginName,
       text: text || '',
-      location: sanitizeLocation(location, where),
+      location: sanitizeLocation(location, where, terminalWidth),
       notes: notesClone,
       detail: stash ? stash.store(detail) : -1,
     })
@@ -1736,13 +1804,14 @@ function sanitizeStringArray(values: any[], property: string): string[] {
   return result
 }
 
-function convertOutputFiles({ path, contents }: protocol.BuildOutputFile): types.OutputFile {
+function convertOutputFiles({ path, contents, hash }: protocol.BuildOutputFile): types.OutputFile {
   // The text is lazily-generated for performance reasons. If no one asks for
   // it, then it never needs to be generated.
   let text: string | null = null
   return {
     path,
     contents,
+    hash,
     get text() {
       // People want to be able to set "contents" and have esbuild automatically
       // derive "text" for them, so grab the contents off of this object instead
