@@ -30,6 +30,7 @@ export interface ServeRequest {
   servedir?: string
   keyfile?: string
   certfile?: string
+  fallback?: string
 }
 
 export interface ServeResponse {
@@ -66,6 +67,7 @@ export interface OnEndResponse {
 export interface BuildOutputFile {
   path: string
   contents: Uint8Array
+  hash: string
 }
 
 export interface PingRequest {
@@ -220,6 +222,7 @@ export interface OnLoadRequest {
   namespace: string
   suffix: string
   pluginData: number
+  with: Record<string, string>
 }
 
 export interface OnLoadResponse {
