@@ -25,10 +25,6 @@ function test(name, backends, fn) {
       promise.then(cancel, cancel)
       promise.then(resolve, reject)
     }),
-
-    // It's ok that the Go WebAssembly runtime uses "setTimeout"
-    sanitizeResources: false,
-    sanitizeOps: false,
   })
 
   for (const backend of backends) {
@@ -41,7 +37,7 @@ function test(name, backends, fn) {
             await fn({ esbuild: esbuildNative, testDir })
             await Deno.remove(testDir, { recursive: true }).catch(() => null)
           } finally {
-            esbuildNative.stop()
+            await esbuildNative.stop()
           }
         })
         break
@@ -55,7 +51,7 @@ function test(name, backends, fn) {
             await fn({ esbuild: esbuildWASM, testDir })
             await Deno.remove(testDir, { recursive: true }).catch(() => null)
           } finally {
-            esbuildWASM.stop()
+            await esbuildWASM.stop()
           }
         })
         break
@@ -69,7 +65,7 @@ function test(name, backends, fn) {
             await fn({ esbuild: esbuildWASM, testDir })
             await Deno.remove(testDir, { recursive: true }).catch(() => null)
           } finally {
-            esbuildWASM.stop()
+            await esbuildWASM.stop()
           }
         })
         break
